//============================================================
// STUDENT NAME: <your name>
// MATRIC NO.  : <matric no.>
// NUS EMAIL   : <your NUS email address>
// COMMENTS TO GRADER:
// <comments to grader, if any>
//
// ============================================================
//
// FILE: convolution.cu


// Include files from C standard library.
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

// Include files to use CUDA runtime API and CUTIL library.
#include <cutil_inline.h>



/////////////////////////////////////////////////////////////////////////////
// CONSTANTS & GLOBAL VARIABLES
/////////////////////////////////////////////////////////////////////////////

// FILTER_WIDTH must be odd, and BLOCK_SIZE >= FILTER_WIDTH.
#define FILTER_WIDTH	469

// Number of CUDA threads per thread block. BLOCK_SIZE >= FILTER_WIDTH.
#define BLOCK_SIZE		512		

// Number of elements in the data.
// Note that DATA_SIZE is always a multiple of BLOCK_SIZE.
#define DATA_SIZE		(500 * BLOCK_SIZE)	

// Number of CUDA thread blocks. 
#define NUM_BLOCKS		( ( (DATA_SIZE) + (BLOCK_SIZE) - 1 ) / (BLOCK_SIZE) )




//===========================================================================
// CUDA Kernel 1. 
// Does not use shared memory.
// Does not care about memory coalesces.
//
// Compute the convolution of the data and the filter.
// Filter width (filterWidth) must be odd, and the filter's
// origin is the center element, that is the element
// filter[ filterWidth/2 ].
//
// For each output element output[i] that does not have enough input 
// data elements in its neighborhood (that is when i < (filterWidth/2) or
// when i >= (dataSize - filterWidth/2)), the output element output[i]
// will have value 0.0.
//
//===========================================================================
__global__ void GPU_Convolve1( const float *data, int dataSize, 
						       const float *filter, int filterWidth,
						       float *output )
{
	int tid = blockIdx.x * blockDim.x + threadIdx.x;

	int filterRadius = filterWidth / 2;

	if ( tid >= filterRadius && tid < (dataSize - filterRadius) )
	{
		float sum = 0.0;

		for ( int k = 0; k < filterWidth; k++ )
			sum += filter[k] * data[ tid - filterRadius + k ];

			// Reading filter[k]: 
			// When k is a certain value, all threads within the half warp 
			// are reading the same filter element filter[k] from the global 
			// memory at the same time. These reads cannot be coalesced, and 
			// will generate 16 32-byte memory transactions (minimum 
			// transaction size). This is very inefficient in memory 
			// bandwidth.
			//
			// Reading data[ tid - filterRadius + k ]:
			// Since filterRadius and k are fixed for all threads within the 
			// half warp, and tid is different for these threads, they will 
			// be reading from different but consecutive elements from data[]. 
			// However, unless the value of (k - filterRadius) is a multiple 
			// of 16, which means the 16 reads are within the same 64-byte 
			// segment, these reads cannot be coalesced, and will generate 16
			// 32-byte memory transactions (minimum transaction size). The 
			// value of (k - filterRadius) is a multiple of 16 only once for 
			// every 16 iterations of k.
		
		output[ tid ] = sum;
	}
	else if ( tid < dataSize )
	{
		output[ tid ] = 0.0;
	}
}



//===========================================================================
// CUDA Kernel 2. 
// Use shared memory.
// Care about memory coalesces.
// Care about shared memory conflicts.
//
// Compute the convolution of the data and the filter.
// Filter width (filterWidth) must be odd, and the filter's
// origin is the center element, that is the element
// filter[ filterWidth/2 ].
//
// For each output element output[i] that does not have enough input 
// data elements in its neighborhood (that is when i < (filterWidth/2) or
// when i >= (dataSize - filterWidth/2)), the output element output[i]
// will have value 0.0. 
//
// Assume that filterWidth <= BLOCK_SIZE.
// Assume that dataSize is a multiple of BLOCK_SIZE.
//
//===========================================================================
__global__ void GPU_Convolve2( const float *data, int dataSize, 
						       const float *filter, int filterWidth,
						       float *output )
{
	__shared__ float filterS[ BLOCK_SIZE ];  
	__shared__ float dataS[ 3 * BLOCK_SIZE ];  

	int tid = blockIdx.x * blockDim.x + threadIdx.x;
	int tx = threadIdx.x;

	int filterRadius = filterWidth / 2;

// First, EVERY thread within the block must cooperate to copy the filter and
// the relevant part of the data array from the global memory to the shared memory. 
// This includes threads whose results are 0 and do not need to use this data at all.


	// Read whole filter from global memory to shared memory.
	if ( tx < filterWidth ) 
		filterS[ tx ] = filter[ tx ];
		// These reads from the global memory are fully coalesced.
		// There is also no shared memory conflict.

	// Read current data tile and the tiles before and after 
	// it into shared memory. 
	// A tile is the same size as a block.

	// Copying the tile before.
	if ( blockIdx.x >= 1 ) 
		dataS[ tx ] = data[ tid - BLOCK_SIZE ];

	// The current tile.
	dataS[ BLOCK_SIZE + tx ] = data[ tid ];						   

	// The tile after.
	if ( blockIdx.x < gridDim.x - 1 ) 
		dataS[ 2 * BLOCK_SIZE + tx ] = data[ tid + BLOCK_SIZE ]; 

		// The reading of the three tiles from the global memory are fully 
		// coalesced because BLOCK_SIZE is a multiple of 16.
		// There is also no shared memory conflict because the threads within
		// each half warp are writing to different and consecutive memory banks.

	__syncthreads();


	if ( tid >= filterRadius && tid < (dataSize - filterRadius) )
	{
		float sum = 0.0;

		for ( int k = 0; k < filterWidth; k++ )
			sum += filterS[k] * dataS[ BLOCK_SIZE + tx - filterRadius + k ];

			// Reading filterS[k]: 
			// When k is a certain value, all threads within the half warp 
			// are reading the same filter element filterS[k] from the shared 
			// memory at the same time. This results in a broadcast
			// to all these threads simultaneously, and thus it is efficient.
			//
			// Reading dataS[ BLOCK_SIZE + tx - filterRadius + k ]:
			// Since BLOCK_SIZE, filterRadius and k are fixed for all threads 
			// within the half warp, and tx is different for these threads, 
			// they will read from different and consecutive banks from the 
			// shared memory. Therefore, there is no shared memory conflict.

		output[ tid ] = sum;
	}
	else if ( tid < dataSize )
	{
		output[ tid ] = 0.0;
	}
}




//===========================================================================
// CPU version.
//
// Compute the convolution of the data and the filter.
// Filter width (filterWidth) must be odd, and the filter's
// origin is the center element, that is the element
// filter[ filterWidth/2 ].
//
// For each output element output[i] that does not have enough input 
// data elements in its neighborhood (that is when i < (filterWidth/2) or
// when i >= (dataSize - filterWidth/2)), the output element output[i]
// will have value 0.0. 
//
//===========================================================================
static void CPU_Convolve( const float *data, int dataSize, 
						  const float *filter, int filterWidth,
						  float *output )
{
	int filterRadius = filterWidth / 2;

	for ( int i = 0; i < dataSize; i++ ) output[i] = 0.0;

	for ( int i = filterRadius; i < (dataSize - filterRadius); i++ ) 
		for ( int k = 0; k < filterWidth; k++ )
			output[i] += filter[k] * data[ i - filterRadius + k ];
}



//===========================================================================
// Returns a random value in the range [min, max] from a uniform distribution.
//===========================================================================
inline static double UniformRandom( double min, double max )  
{
    return ( ((double)rand()) / RAND_MAX ) * (max - min) + min;
}



//===========================================================================
// Generates a set of random floating-point numbers in the range [min,max] 
// and put them in the the array A.
//===========================================================================
static void GenerateRandomArray( float *A, int numElems, float min, float max )
{
	for ( int i = 0; i < numElems; i++ )
		A[i] = (float) UniformRandom( min, max );
}



//===========================================================================
// Return true iff all corresponding elements in the float arrays A and B 
// are approximately equal (i.e. the absolute difference is within the
// given epsilon).
//===========================================================================
static bool FloatArrayEqual( const float *A, const float *B, int numElems, float epsilon )
{
    for ( int i = 0; i < numElems; i++ )
        if ( fabs( A[i] - B[i] ) > epsilon ) return false;

    return true;
}



//===========================================================================
// The main function
//===========================================================================
int main(int argc, char** argv)
{
	// Select a device (GPU) for running the kernel.
    if( cutCheckCmdLineFlag( argc, (const char**) argv, "device" ) )
        cutilDeviceInit( argc, argv );
    else
        cudaSetDevice( cutGetMaxGflopsDeviceId() );

    // Set random seed for rand().
    srand( 927 );

    // Create a timer.
    unsigned int timer = 0;
    cutilCheckError( cutCreateTimer( &timer ) );

//---------------------------------------------------------------------------
// Allocate memory and generate test data.
//---------------------------------------------------------------------------

    // Allocate host memory for filter, input data and result arrays.
	float *h_filter = (float *) malloc( FILTER_WIDTH * sizeof(float) );
	float *h_data = (float *) malloc( DATA_SIZE * sizeof(float) );
	float *h_output = (float *) malloc( DATA_SIZE * sizeof(float) );

	// Allocate host memory for receiving results from the GPU.
	float *d2h_output1 = (float *) malloc( DATA_SIZE * sizeof(float) );
	float *d2h_output2 = (float *) malloc( DATA_SIZE * sizeof(float) );

    // Allocate device memory.
	float *d_filter, *d_data, *d_output;
	cutilSafeCall( cudaMalloc( (void**) &d_filter, FILTER_WIDTH * sizeof(float) ) );
	cutilSafeCall( cudaMalloc( (void**) &d_data, DATA_SIZE * sizeof(float) ) );
	cutilSafeCall( cudaMalloc( (void**) &d_output, DATA_SIZE * sizeof(float) ) );

 	// Fill the host filter and data arrays with random floating-point numbers.
	GenerateRandomArray( h_filter, FILTER_WIDTH, 0.0, 1.0 );
	GenerateRandomArray( h_data, DATA_SIZE, 1.0, 5.0 );

//---------------------------------------------------------------------------
// Print some program parameter values.
//---------------------------------------------------------------------------
	printf( "Filter width = %d\n", FILTER_WIDTH );
	printf( "Data size = %d\n", DATA_SIZE );
	printf( "Thread block size = %d\n", BLOCK_SIZE );
	printf( "Number of thread blocks = %d\n", NUM_BLOCKS );
    printf( "\n\n" );


//---------------------------------------------------------------------------
// Perform computation on CPU.
//---------------------------------------------------------------------------
    printf( "CPU COMPUTATION:\n" );

	// Reset and start timer.
	cutilCheckError( cutResetTimer( timer ) );
	cutilCheckError( cutStartTimer( timer ) );

	// Compute on CPU.
	CPU_Convolve( h_data, DATA_SIZE, h_filter, FILTER_WIDTH, h_output );

	// Stop timer.
	cutilCheckError( cutStopTimer( timer ) );

    printf( "Processing time = %.3f ms\n", cutGetTimerValue(timer) );

    // Print some results.
    printf( "First element  = %.8f\n", h_output[0] );
    printf( "Middle element = %.8f\n", h_output[ DATA_SIZE / 2 ] );
    printf( "Last element   = %.8f\n", h_output[ DATA_SIZE - 1 ] );
    printf( "\n\n" );


//---------------------------------------------------------------------------
// Perform computation on GPU using Kernel 1 (not using shared memory).
//---------------------------------------------------------------------------
	printf( "GPU COMPUTATION 1 (not using Shared Memory):\n" );

	// Reset and start timer.
	cutilCheckError( cutResetTimer( timer ) );
	cutilCheckError( cutStartTimer( timer ) );

	// Copy host memory to device.
    cutilSafeCall( cudaMemcpy( d_filter, h_filter, FILTER_WIDTH * sizeof(float), cudaMemcpyHostToDevice ) );
    cutilSafeCall( cudaMemcpy( d_data, h_data, DATA_SIZE * sizeof(float), cudaMemcpyHostToDevice ) );

	// Clear the output array in device memory.
    cutilSafeCall( cudaMemset( d_output, 0, DATA_SIZE * sizeof(float) ) );

	// Execute the kernel.
	GPU_Convolve1 <<<NUM_BLOCKS, BLOCK_SIZE>>> ( d_data, DATA_SIZE, d_filter, FILTER_WIDTH, d_output );

    // Check if kernel execution generated any error.
    cutilCheckMsg( "Kernel execution failed" );

	// Copy result from device memory to host.
    cutilSafeCall( cudaMemcpy( d2h_output1, d_output, DATA_SIZE * sizeof(float), cudaMemcpyDeviceToHost ) );

	// Stop timer.
	cutilCheckError( cutStopTimer( timer ) );

    printf( "Processing time = %.3f ms\n", cutGetTimerValue(timer) );

    // Print some results.
    printf( "First element  = %.8f\n", d2h_output1[0] );
    printf( "Middle element = %.8f\n", d2h_output1[ DATA_SIZE / 2 ] );
    printf( "Last element   = %.8f\n", d2h_output1[ DATA_SIZE - 1 ] );

	// Check result with reference result computed by CPU.
	bool equal1 = FloatArrayEqual( h_output, d2h_output1, DATA_SIZE, 0.001f );
	printf( "Verify GPU result... %s\n", (equal1)? "PASS" : "FAIL" );
    printf( "\n\n" );


//---------------------------------------------------------------------------
// Perform computation on GPU using Kernel 2 (using shared memory).
//---------------------------------------------------------------------------
	printf( "GPU COMPUTATION 2 (using Shared Memory):\n" );

	// Reset and start timer.
	cutilCheckError( cutResetTimer( timer ) );
	cutilCheckError( cutStartTimer( timer ) );

	// Copy host memory to device.
    cutilSafeCall( cudaMemcpy( d_filter, h_filter, FILTER_WIDTH * sizeof(float), cudaMemcpyHostToDevice ) );
    cutilSafeCall( cudaMemcpy( d_data, h_data, DATA_SIZE * sizeof(float), cudaMemcpyHostToDevice ) );

	// Clear the output array in device memory.
    cutilSafeCall( cudaMemset( d_output, 0, DATA_SIZE * sizeof(float) ) );

	// Execute the kernel.
	GPU_Convolve2 <<<NUM_BLOCKS, BLOCK_SIZE>>> ( d_data, DATA_SIZE, d_filter, FILTER_WIDTH, d_output );

    // Check if kernel execution generated any error.
    cutilCheckMsg( "Kernel execution failed" );

	// Copy result from device memory to host.
    cutilSafeCall( cudaMemcpy( d2h_output2, d_output, DATA_SIZE * sizeof(float), cudaMemcpyDeviceToHost ) );

	// Stop timer.
	cutilCheckError( cutStopTimer( timer ) );

    printf( "Processing time = %.3f ms\n", cutGetTimerValue(timer) );

    // Print some results.
    printf( "First element  = %.8f\n", d2h_output2[0] );
    printf( "Middle element = %.8f\n", d2h_output2[ DATA_SIZE / 2 ] );
    printf( "Last element   = %.8f\n", d2h_output2[ DATA_SIZE - 1 ] );

	// Check result with reference result computed by CPU.
	bool equal2 = FloatArrayEqual( h_output, d2h_output2, DATA_SIZE, 0.001f );
	printf( "Verify GPU result... %s\n", (equal2)? "PASS" : "FAIL" );
    printf( "\n\n" );


//---------------------------------------------------------------------------
// Clean up.
//---------------------------------------------------------------------------
	// Destroy timer.
	cutilCheckError( cutDeleteTimer( timer ) );

    // Free up memory.
    free( h_filter );
    free( h_data );
    free( h_output );
    free( d2h_output1 );
    free( d2h_output2 );
    cutilSafeCall( cudaFree( d_filter ) );
    cutilSafeCall( cudaFree( d_data ) );
    cutilSafeCall( cudaFree( d_output ) );

	// Prevents the console window from closing immediately before program terminates.
    cutilExit( argc, argv );  
}
