//============================================================
// STUDENT NAME: <your name>
// MATRIC NO.  : <matric no.>
// NUS EMAIL   : <your NUS email address>
// COMMENTS TO GRADER:
// <comments to grader, if any>
//
// ============================================================
//
// FILE: prefixsums.cpp


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <GL/glew.h>
#include <GL/glut.h>

#include "shader_util.h"
#include "timing.h"


/////////////////////////////////////////////////////////////////////////////
// CONSTANTS & GLOBAL VARIABLES
/////////////////////////////////////////////////////////////////////////////

// Size of input texture.
static const int texWidth  = 1856;
static const int texHeight = 1934;

// Size of the input array of elements.
static const int numInputElems = texWidth * texHeight;

// Shaders' filenames.
static const char vertShaderFilename[] = "prefixsums.vert";
static const char fragShaderFilename[] = "prefixsums.frag";

// Shader program object.
static GLuint shaderProg = 0;

// GLUT window handle.
static GLuint glutWindowHandle;



/////////////////////////////////////////////////////////////////////////////
// Compute the inclusive all-prefix-sums of the input array on the CPU.
// Note that the memory for outputArray must be already allocated.
/////////////////////////////////////////////////////////////////////////////
static void CPU_PrefixSums( const float inputArray[], float outputArray[], int numElems )
{
	// Assuming there is at least one element in the input array.
	outputArray[0] = inputArray[0];

    for ( int i = 1; i < numElems; i++ ) 
		outputArray[i] = outputArray[i-1] + inputArray[i];
}



/////////////////////////////////////////////////////////////////////////////
// Check framebuffer status.
// Modified from the sample code provided in the 
// GL_EXT_framebuffer_object extension sepcifications.
/////////////////////////////////////////////////////////////////////////////
static bool CheckFramebufferStatus() 
{
    GLenum status = (GLenum) glCheckFramebufferStatusEXT( GL_FRAMEBUFFER_EXT );
    switch( status ) 
    {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            return true;
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
            fprintf( stderr, "Framebuffer incomplete, incomplete attachment\n" );
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
            fprintf( stderr, "Framebuffer incomplete, missing attachment\n" );
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
            fprintf( stderr, "Framebuffer incomplete, attached images must have same dimensions\n" );
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
            fprintf( stderr, "Framebuffer incomplete, attached images must have same format\n" );
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
            fprintf( stderr, "Framebuffer incomplete, missing draw buffer\n" );
            return false;
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
            fprintf( stderr, "Framebuffer incomplete, missing read buffer\n" );
            return false;
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
            fprintf( stderr, "Unsupported framebuffer format\n" );
            return false;
    }
    return false;
}



/////////////////////////////////////////////////////////////////////////////
// Compute the inclusive all-prefix-sums of the input array on the GPU.
// Note that the memory for outputArray must be already allocated.
/////////////////////////////////////////////////////////////////////////////
static void GPU_PrefixSums( const float inputArray[], float outputArray[], int texWidth, int texHeight )
{
//-----------------------------------------------------------------------------
// Create two floating-point textures. 
// Use texture rectangle and texture internal format GL_ALPHA32F_ARB.
//-----------------------------------------------------------------------------
	// Texture A.
    glActiveTexture( GL_TEXTURE0 );
    GLuint texA;
    glGenTextures( 1, &texA );
    glBindTexture( GL_TEXTURE_RECTANGLE_ARB, texA );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, GL_ALPHA32F_ARB,
                  texWidth, texHeight, 0, GL_ALPHA, GL_FLOAT, inputArray );
    printOpenGLError();

	// Texture B.
    glActiveTexture( GL_TEXTURE1 );
    GLuint texB;
    glGenTextures( 1, &texB );
    glBindTexture( GL_TEXTURE_RECTANGLE_ARB, texB );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP );
    glTexParameteri( GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP );
    glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, GL_ALPHA32F_ARB,
                  texWidth, texHeight, 0, GL_ALPHA, GL_FLOAT, NULL );
    printOpenGLError();

//-----------------------------------------------------------------------------
// Attach the two textures to a FBO.
//-----------------------------------------------------------------------------
    GLuint fbo;
    glGenFramebuffersEXT( 1, &fbo ); 
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, fbo );

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, 
                               GL_TEXTURE_RECTANGLE_ARB, texA, 0 );
	CheckFramebufferStatus();
    printOpenGLError();

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, 
                               GL_TEXTURE_RECTANGLE_ARB, texB, 0 );
    CheckFramebufferStatus();
    printOpenGLError();

//-----------------------------------------------------------------------------
// Deploy user-defined shaders.
//-----------------------------------------------------------------------------
    glUseProgram( shaderProg );

	GLint uniInputTexLoc = glGetUniformLocation( shaderProg, "InputTex" );
	GLint uniTexWidthLoc = glGetUniformLocation( shaderProg, "TexWidth" );
	GLint uniTexHeightLoc = glGetUniformLocation( shaderProg, "TexHeight" );
	GLint uniPassCountLoc = glGetUniformLocation( shaderProg, "PassCount" );
	GLint uniStrideLoc = glGetUniformLocation( shaderProg, "Stride" );

	// These uniform variables' values do not change in all rendering passes.
	glUniform1i( uniTexWidthLoc, texWidth );
	glUniform1i( uniTexHeightLoc, texHeight );


//-----------------------------------------------------------------------------
// Set some OpenGL states and projection.
//-----------------------------------------------------------------------------
    glDisable( GL_DITHER );
    glDisable( GL_DEPTH_TEST );
    glDisable( GL_BLEND );
    glDisable( GL_ALPHA_TEST );
    glDisable( GL_COLOR_LOGIC_OP );
    glDisable( GL_SCISSOR_TEST );
    glDisable( GL_STENCIL_TEST );
    glPolygonMode( GL_FRONT, GL_FILL );

	// Set up projection, modelview matrices and viewport.
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluOrtho2D( 0, texWidth, 0, texHeight );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glViewport( 0, 0, texWidth, texHeight );
    
//-----------------------------------------------------------------------------
// Perform the rendering passes.
// Use the ping-pong technique.
//-----------------------------------------------------------------------------

	GLint inputTextureUnit = 0;
	GLenum outputFBOAttachement = GL_COLOR_ATTACHMENT1_EXT;

	int passCount = 1;
	int stride = 1;

	glFinish();  // For timing purpose only.
    double starttime = GetCurrRealTime_A();

	while ( true )
	{
	//---------------------------------------------------------
	// Perform a rendering pass.
	//---------------------------------------------------------
		glDrawBuffer( outputFBOAttachement );

		glUniform1i( uniInputTexLoc, inputTextureUnit );
		glUniform1i( uniPassCountLoc, passCount );
		glUniform1i( uniStrideLoc, stride );

		// Draw a filled quad with size equal to the output size.
		glBegin( GL_QUADS );
			glVertex2f( 0, 0 );
			glVertex2f( texWidth, 0 );
			glVertex2f( texWidth, texHeight );
			glVertex2f( 0, texHeight );
		glEnd();
		CheckFramebufferStatus();
		printOpenGLError();

		passCount++;
		stride *= 2;

	//---------------------------------------------------------
	// Break out of while-loop if final rendering pass is done.
	//---------------------------------------------------------
		if ( stride >= texWidth * texHeight ) break;

	//---------------------------------------------------------
	// Prepare for next rendering pass.
	//---------------------------------------------------------
		// Swap input and output textures. 
		if ( inputTextureUnit == 0 ) 
			inputTextureUnit = 1;
		else
			inputTextureUnit = 0;

		if ( outputFBOAttachement == GL_COLOR_ATTACHMENT1_EXT )
			outputFBOAttachement = GL_COLOR_ATTACHMENT0_EXT;
		else
			outputFBOAttachement = GL_COLOR_ATTACHMENT1_EXT;
	}

	glFinish();  // For timing purpose only.
    double endtime = GetCurrRealTime_A();
	printf( "Time elapsed (exclude CPU-GPU data transfer) = %.4f sec\n", endtime - starttime );


//-----------------------------------------------------------------------------
// Read output buffer/texture to CPU memory.
//-----------------------------------------------------------------------------
    glPixelStorei( GL_PACK_ALIGNMENT, 1 );
    glReadBuffer( outputFBOAttachement );
    glReadPixels( 0, 0, texWidth, texHeight, GL_ALPHA, GL_FLOAT, outputArray );
    printOpenGLError();

//-----------------------------------------------------------------------------
// Clean up.
//-----------------------------------------------------------------------------
    glDeleteFramebuffersEXT( 1, &fbo );
    glDeleteTextures( 1, &texA );
    glDeleteTextures( 1, &texB );
}




/////////////////////////////////////////////////////////////////////////////
// Create a OpenGL rendering context. 
// Check for OpenGL 2.0 and the necessary OpenGL extensions. 
// Read in the shaders from files to create a shader program object.
/////////////////////////////////////////////////////////////////////////////
static void PrepareGPUExecution( int argc, char** argv )
{
// Initialize GLUT.
    glutInit( &argc, argv );

    // This creates a OpenGL rendering context so that
    // we can start to issue OpenGL commands after this.
    glutWindowHandle = glutCreateWindow( "" );  
 
// Initialize GLEW.
    GLenum err = glewInit();
    if ( err != GLEW_OK )
    {
        fprintf( stderr, "Error: %s.\n", glewGetErrorString( err ) );
        char ch; scanf( "%c", &ch ); // Prevents the console window from closing.
        exit( 1 );
    }

// Make sure OpenGL 2.0 is supported.
    if ( !GLEW_VERSION_2_0 )
    {
        fprintf( stderr, "Error: OpenGL 2.0 is not supported.\n" );
        char ch; scanf( "%c", &ch ); // Prevents the console window from closing.
        exit( 1 );
    }

// Make sure necessary OpenGL extensions are supported.
    if ( !GLEW_ARB_texture_float || 
         !GLEW_EXT_framebuffer_object || 
         !GLEW_ARB_texture_rectangle ||
		 !GLEW_EXT_gpu_shader4 )
    {
        fprintf( stderr, "Error: Some necessary OpenGL extensions are not supported.\n" );
        char ch; scanf( "%c", &ch ); // Prevents the console window from closing.
        exit( 1 );
    }

// Create shader program object.
    shaderProg = makeShaderProgramFromFiles( vertShaderFilename, fragShaderFilename, NULL );
    if ( shaderProg == 0 )
    {
        fprintf( stderr, "Error: Cannot create shader program object.\n" );
        char ch; scanf( "%c", &ch ); // Prevents the console window from closing.
        exit( 1 );
    }
}




/////////////////////////////////////////////////////////////////////////////
// Generate an array of random numbers.
// Note that the memory for outputArray must be already allocated.
/////////////////////////////////////////////////////////////////////////////
static void GenerateInputArray( float outputArray[], int numElems )
{
    for ( int i = 0; i < numElems; i++ )
        outputArray[i] = (float) ( rand() % 4 + 1 );
}




/////////////////////////////////////////////////////////////////////////////
// Return true iff all corresponding elements in arrays A and B are 
// approximately equal (i.e. the absolute difference is within the
// given threshold).
/////////////////////////////////////////////////////////////////////////////
static bool ArrayEqual( const float A[], const float B[], int numElems, float threshold )
{
    for ( int i = 0; i < numElems; i++ )
        if ( fabs( A[i] - B[i] ) > threshold ) return false;

    return true;
}




/////////////////////////////////////////////////////////////////////////////
// The main function.
/////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv )
{
    double starttime, endtime;

    srand( 927 ); // To use same set of random numbers every run.

//-----------------------------------------------------------------------------
// Allocate memory for input and output arrays.
//-----------------------------------------------------------------------------
	float *inputArray = (float*) malloc( numInputElems * sizeof(float) );
	float *CPU_outputArray = (float*) malloc( numInputElems * sizeof(float) );
	float *GPU_outputArray = (float*) malloc( numInputElems * sizeof(float) );

	if ( inputArray == NULL || CPU_outputArray == NULL || GPU_outputArray == NULL )
    {
        fprintf( stderr, "Error: Not enough memory.\n" );
        char ch; scanf( "%c", &ch ); // Prevents the console window from closing.
        exit( 1 );
    }

//-----------------------------------------------------------------------------
// Generate elements in input array.
//-----------------------------------------------------------------------------
    GenerateInputArray( inputArray, numInputElems );


//-----------------------------------------------------------------------------
// Perform computation on CPU.
//-----------------------------------------------------------------------------
    printf( "CPU COMPUTATION:\n" );
    starttime = GetCurrRealTime_A();

	CPU_PrefixSums( inputArray, CPU_outputArray, numInputElems );

    endtime = GetCurrRealTime_A();
    printf( "Time elapsed = %.4f sec\n", endtime - starttime );

    // Print some results.
    printf( "Last element of output array = %.8f\n", CPU_outputArray[numInputElems-1] );
    printf( "\n\n" );


//-----------------------------------------------------------------------------
// Perform computation on GPU.
//-----------------------------------------------------------------------------
    printf( "GPU COMPUTATION:\n" );

	PrepareGPUExecution( argc, argv );
	
    starttime = GetCurrRealTime_A();

	GPU_PrefixSums( inputArray, GPU_outputArray, texWidth, texHeight );

    endtime = GetCurrRealTime_A();
    printf( "Time elapsed = %.4f sec\n", endtime - starttime );

    // Print some results.
    printf( "Last element of output array = %.8f\n", GPU_outputArray[numInputElems-1] );
    printf( "\n\n" );


//-----------------------------------------------------------------------------
// Verify results.
//-----------------------------------------------------------------------------
    if ( ArrayEqual( CPU_outputArray, GPU_outputArray, numInputElems, 0.01f ) )
        printf( "GPU result EQUAL to CPU result.\n" );
    else
        printf( "GPU result NOT EQUAL to CPU result.\n" );


//-----------------------------------------------------------------------------
// Clean up.
//-----------------------------------------------------------------------------	
	free( inputArray );
	free( CPU_outputArray );
	free( GPU_outputArray );

    char ch; scanf( "%c", &ch ); // Prevents the console window from closing.
    return 0;
}
