#include <stdlib.h>
#include <stdio.h>
#include <FreeImage.h>
#include "image_io.h"


/////////////////////////////////////////////////////////////////////////////
// Deallocate the memory allocated to (*imageData) returned by 
// the function ReadImageFile().
// (*imageData) will be set to NULL.
/////////////////////////////////////////////////////////////////////////////
void DeallocateImageData( uchar **imageData )
{
    free( *imageData );
    (*imageData) = NULL;
}


/////////////////////////////////////////////////////////////////////////////
// Read a image from the input filename. 
// Returns 1 if successful or 0 if unsuccessful.
// The returned image data will be pointed to by (*imageData).
// The image width, image height, and number of components (color channels) 
// per pixel will be returned in (*imageWidth), (*imageHeight),
// and (*numComponents).
// The value of (*numComponents) can be 1, 2, 3 or 4.
// The returned image data is always packed tightly with red, green, blue,
// and alpha arranged from lower to higher memory addresses. 
// Each color channel take one byte.
/////////////////////////////////////////////////////////////////////////////
int ReadImageFile( const char *filename, uchar **imageData,
                   int *imageWidth, int *imageHeight, int *numComponents )
{

// Call this ONLY when linking with FreeImage as a static library.
#ifdef FREEIMAGE_LIB
    FreeImage_Initialise();
#endif

// Determine image format.
    FREE_IMAGE_FORMAT fif = FreeImage_GetFileType( filename, 0 );
    if( fif == FIF_UNKNOWN ) fif = FreeImage_GetFIFFromFilename( filename );
    if( fif == FIF_UNKNOWN )
    {
        printf( "Error: Cannot determine image format of %s.\n", filename );
        return 0;
    }

// Read image data from file.
    FIBITMAP *dib = NULL;
    if( FreeImage_FIFSupportsReading( fif ) )
        dib = FreeImage_Load( fif, filename );

    if( !dib )
    {
        printf( "Error: Cannot read image file %s.\n", filename );
        return 0;
    }

// Check image type.
    FREE_IMAGE_TYPE fit = FreeImage_GetImageType( dib );
    if ( fit != FIT_BITMAP )
    {
        FreeImage_Unload( dib );
        printf( "Error: Only 8-bits-per-component standard bitmap is supported.\n" );
        return 0;
    }

// Check bits per pixel.
    int bits_per_pixel = FreeImage_GetBPP( dib );
    if ( bits_per_pixel != 8 && bits_per_pixel != 16 && bits_per_pixel != 24 && bits_per_pixel != 32 )
    {
        FreeImage_Unload( dib );
        printf( "Error: Only 8, 16, 24, 32 bits per pixel are supported.\n" );
        return 0;
    }

    int _numComponents = bits_per_pixel / 8;
    int _imageWidth = FreeImage_GetWidth( dib );
    int _imageHeight = FreeImage_GetHeight( dib );
    uchar *_imageData = (uchar *) malloc( _imageWidth * _imageHeight * _numComponents );
    if ( _imageData == NULL )
    {
        FreeImage_Unload( dib );
        printf( "Error: Not enough memory.\n" );
        return 0;
    }

    int imageDataCount = 0;

    if ( _numComponents == 1 )
    {
        for( int y = 0; y < _imageHeight; y++ )
        {
            BYTE *dibData = FreeImage_GetScanLine( dib, y );

            for( int x = 0; x < _imageWidth; x++) 
            {
                _imageData[imageDataCount++] = dibData[x];
            }
        }
    }
    else if ( _numComponents == 2 )
    {
        for( int y = 0; y < _imageHeight; y++ )
        {
            BYTE *dibData = FreeImage_GetScanLine( dib, y );

            for( int x = 0; x < _imageWidth; x++) 
            {
                _imageData[imageDataCount++] = dibData[0];
                _imageData[imageDataCount++] = dibData[1];
                dibData += _numComponents;
            }
        }
    }
    else if ( _numComponents == 3 )
    {
        for( int y = 0; y < _imageHeight; y++ )
        {
            BYTE *dibData = FreeImage_GetScanLine( dib, y );

            for( int x = 0; x < _imageWidth; x++) 
            {
                _imageData[imageDataCount++] = dibData[FI_RGBA_RED];
                _imageData[imageDataCount++] = dibData[FI_RGBA_GREEN];
                _imageData[imageDataCount++] = dibData[FI_RGBA_BLUE];
                dibData += _numComponents;
            }
        }
    }
    else if ( _numComponents == 4 )
    {
        for( int y = 0; y < _imageHeight; y++ )
        {
            BYTE *dibData = FreeImage_GetScanLine( dib, y );

            for( int x = 0; x < _imageWidth; x++) 
            {
                _imageData[imageDataCount++] = dibData[FI_RGBA_RED];
                _imageData[imageDataCount++] = dibData[FI_RGBA_GREEN];
                _imageData[imageDataCount++] = dibData[FI_RGBA_BLUE];
                _imageData[imageDataCount++] = dibData[FI_RGBA_ALPHA];
                dibData += _numComponents;
            }
        }
    }

    FreeImage_Unload( dib );

// Call this ONLY when linking with FreeImage as a static library.
#ifdef FREEIMAGE_LIB
    FreeImage_DeInitialise();
#endif

    (*numComponents) = _numComponents;
    (*imageWidth) = _imageWidth;
    (*imageHeight) = _imageHeight;
    (*imageData) = _imageData;
    return 1; 
}
